#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "symbol_table.h"
#include "clauses.h"
#include "prover.h"
#include "uthash.h"

#define CNF_INTRO 19
#define CNF_EOL 3
#define SPACE 2
#define SAT_CMD_SIZE 46
#define SAT_CMD "' | ./minisat/core/minisat_static -verb=0"
#define ECHO_CMD_SIZE 14
#define ECHO_CMD "/bin/echo -e '"
#define SPEC_CMD 10

extern int numclause;
extern hml_pclauses *l_sos;

extern tclause *get_clause(int clause);
extern void print_clause(tclause *clause);
extern int classify_clause(literalslist *l);
extern int size_clause(literalslist *l);
extern int max_lit_clause(literalslist *l);
extern char *itoa(int val);
extern void insert_clause_prop(tclause **clause, hml_pclauses **set);
void insert_clause_set (tclause **clause, int type, int set);
extern literalslist *free_disjunction(literalslist *dis);
extern void print_literal(int lit);
extern void print_disjunction(literalslist *s);
extern void do_gen3(int clause);
extern void print_pclauses (hml_pclauses *set);
extern int same(tclause *c1, tclause *c2);

extern literalslist *simplify_clause(literalslist *l);
extern clauseslist *delete_clauses_list (clauseslist *l);

extern hml_pclauses *l_sos;
extern hml_pclauses *l_usable;

void get_cnfclauses(int ml, clauseslist *c);
void learn_clauses_at_modal_level(int ml);
void learn_clauses_at_modal_level_with_clauses(char* propclauses, int ml);

char *cmd = NULL;

/* Second level by agent */
/* @TODO consider moving all modal literals to a single hash */
typedef struct hag_cnfliterals {
    int ag;
    /* @TODO move list of cnf literals to a heap */
    literalslist *negatives;
    literalslist *positives;
    char *clause;
    UT_hash_handle hag;
} hag_cnfliterals;

/* First level by modal level */
typedef struct hml_cnfliterals {
    int ml;
    hag_cnfliterals *ag;
    UT_hash_handle hml;
} hml_cnfliterals;

/* All propositional clauses in cnf by modal level */
typedef struct hml_propclauses {
    int ml;
    char *clauses;
    UT_hash_handle hml;
} hml_propclauses;

hml_cnfliterals *modalliterals = NULL;
hml_propclauses *propliterals = NULL;

hml_pclauses *learnt_clauses = NULL;

int largest = 0;
int numClauses = 0;

int sign(int a) { return a >= 0 ? 1 : -1; }

void print_cnfliterals () {
    hml_cnfliterals *aux = NULL;
    hag_cnfliterals *ag = NULL;

    for (aux = modalliterals; aux != NULL; aux = aux->hml.next) {
        printf("[%d](+) ", aux->ml);
        for(ag = aux->ag; ag != NULL; ag = ag->hag.next) {
            literalslist *list = ag->positives;
            int l = list->literal;
            while(list != NULL) {
                printf("%d (", l);
                print_literal((MAX_INT - abs(l) + 1)*sign(l));
                printf(") ");
                list = list->next;
            }
            printf("\n");
            list = ag->negatives;
            l = list->literal;
            while(list != NULL) {
                printf("%d (", l);
                print_literal((MAX_INT - abs(l) + 1)*sign(l));
                printf(") ");
                list = list->next;
            }
            printf("\n");
        }
    }
}

void add_modal_cnfliterals(int type, int ml, int ag, int cnflit) {
    hml_cnfliterals *hml = NULL;
    hag_cnfliterals *hag = NULL;
    literalslist *cnf_lits = NULL;
    char *string_lit;
    
    HASH_FIND(hml, modalliterals, &(ml), sizeof(int), hml);
    if(hml == NULL) {
        hml_cnfliterals *new = malloc(sizeof(hml_cnfliterals));
        new->ml = ml;
        new->ag = NULL;
        HASH_ADD(hml, modalliterals, ml, sizeof(int), new);
        hml = new;
    }

    HASH_FIND(hag, hml->ag, &(ag), sizeof(int), hag);
    if(hag == NULL) {
        hag_cnfliterals *new = malloc(sizeof(hag_cnfliterals));
        new->ag = ag;
        new->negatives = new->positives = NULL;
        new->clause = malloc(sizeof(char));
        strcpy(new->clause, "");
        HASH_ADD(hag, hml->ag, ag, sizeof(int), new);
        hag = new;
    }

    cnf_lits = malloc(sizeof(literalslist));
    cnf_lits->literal = cnflit;
    if(type == MPOSITIVE) {
        cnf_lits->next = hag->positives;
        hag->positives = cnf_lits;
        string_lit = itoa(cnflit);
        hag->clause = realloc(hag->clause, strlen(hag->clause) + strlen(string_lit) + CNF_EOL + SPACE);
        strcat(hag->clause, string_lit);
        strcat(hag->clause, " 0\n");
        free(string_lit);
    }

    else { //if(type == MNEGATIVE)
        cnf_lits->next = hag->negatives;
        hag->negatives = cnf_lits;
    }

}

void compute_cnf_prop_clauses_at_modal_level (hmaxclauses *set, int ml) {
    hmaxclauses *hmax = NULL;
    hsizeclauses *hsc = NULL;
    clauseslist *clauses = NULL;

    for (hmax = set; hmax != NULL; hmax = hmax->hmax.next) {
        for (hsc = hmax->size; hsc != NULL; hsc = hsc->hsize.next) {
            clauses = hsc->list;
            get_cnfclauses(ml, clauses);
        }
    }
}

void compute_cnf_prop_clauses (hml_pclauses *set) {
    if (set == NULL) return;

    hml_pclauses *hml = NULL;
    hmaxclauses *hmax = NULL;
    hsizeclauses *hsc = NULL;
    clauseslist *clauses = NULL;

    for (hml = set; hml != NULL; hml = hml->hml.next) {
        /*compute_cnf_prop_clauses_at_modal_level(hml->max, hml->ml);*/
        for (hmax = hml->max; hmax != NULL; hmax = hmax->hmax.next) {
            for (hsc = hmax->size; hsc != NULL; hsc = hsc->hsize.next) {
                clauses = hsc->list;
                get_cnfclauses(hml->ml, clauses);
            }
        }
    }
}

void compute_cnf_modal_clauses(hml_mclauses *set) {

    if (set == NULL) return;

    hml_mclauses *hml_iterator = NULL;
    hagentclauses *hag_iterator = NULL;
    htypeclauses *htc_iterator = NULL;
    hmaxlitclauses *hmaxlit = NULL;
    clauseslist *clauses_iterator = NULL;
    tclause *current_clause = NULL;
    int ml;
    int ag;
    int cnflit, lit;

    for (hml_iterator = set; hml_iterator != NULL; hml_iterator = hml_iterator->hml.next) {
        ml = hml_iterator->ml;
        for (hag_iterator = hml_iterator->id; hag_iterator != NULL; hag_iterator = hag_iterator->hid.next) {
            ag = hag_iterator->id;
            for (htc_iterator = hag_iterator->type; htc_iterator != NULL; htc_iterator = htc_iterator->htype.next) {
                for (hmaxlit = htc_iterator->lit; hmaxlit != NULL; hmaxlit = hmaxlit->hmax.next) {
                    clauses_iterator = hmaxlit->list;
                    while (clauses_iterator != NULL) {
                        current_clause = get_clause(clauses_iterator->clause_number);
                        lit = current_clause->right->literal;
                        cnflit = (MAX_INT - abs(lit) + 1) * sign(lit);
                        if (htc_iterator->type == MPOSITIVE) {
                            add_modal_cnfliterals(MPOSITIVE, ml, ag, cnflit);
                        }
                        else /*if (htc_iterator->type==MNEGATIVE) */ {
                            add_modal_cnfliterals(MNEGATIVE, ml, ag, cnflit);
                        }
                        clauses_iterator = clauses_iterator->next;
                    }
                }
            }
        }
    }
}

void get_cnfclauses(int ml, clauseslist *c) {

    hml_propclauses *hml = NULL;
    literalslist *literals = NULL;
    char *clauses;
    char *lit;
    int d;

    HASH_FIND(hml, propliterals, &ml, sizeof(int), hml);
    if(hml == NULL) {
        hml_propclauses *new = malloc(sizeof(hml_propclauses));
        new->ml = ml;
        new->clauses = malloc((size_t) CNF_INTRO);
        sprintf(new->clauses, "c input from ksp\n");
        HASH_ADD(hml, propliterals, ml, sizeof(int), new);
        hml = new;
    }

    clauses = hml->clauses;

    while (c != NULL) {
        tclause *clause = get_clause(c->clause_number);
        literals = clause->right;

        while (literals != NULL) {
            d = literals->literal;
            d = (MAX_INT - abs(d) + 1) * sign(d);
            lit = itoa(d);
            clauses = (char*) realloc(clauses, strlen(clauses) + strlen(lit) + SPACE);
            strcat(clauses, lit);
            strcat(clauses, " ");
            literals = literals->next;
            free(lit);
            lit = NULL;
        }

        clauses = (char*) realloc(clauses, strlen(clauses) + (size_t)CNF_EOL);
        strcat(clauses, "0\n");
        c = c->next;
    }
    hml->clauses = clauses;
}

void build_command(char *prop, char *mod) {

    cmd = malloc(ECHO_CMD_SIZE + strlen(prop) + strlen(mod) + SAT_CMD_SIZE + 5);
    sprintf(cmd, "%s", ECHO_CMD);
    strcat(cmd, prop);
    strcat(cmd, mod);
    strcat(cmd, SAT_CMD);
}

void destroy_propclauses() {
    if (propliterals == NULL) return;
    hml_propclauses *curr = NULL, *next = NULL;

    HASH_ITER(hml, propliterals, curr, next) {
        free(curr->clauses);
        curr->clauses = NULL;
        HASH_DELETE(hml,propliterals,curr);
        free(curr);
    }

    HASH_CLEAR(hml, propliterals);
    free(propliterals);
    propliterals = NULL;
}

void free_cnf_modal_hash(hml_cnfliterals **hash) {
    hml_cnfliterals *hml_curr = NULL, *hml_next = NULL;
    hag_cnfliterals *hag_curr = NULL, *hag_next = NULL;

    HASH_ITER(hml, *hash, hml_curr, hml_next) {
        HASH_ITER(hag, hml_curr->ag, hag_curr, hag_next) {
            hag_curr->negatives = free_disjunction(hag_curr->negatives);
            hag_curr->positives = free_disjunction(hag_curr->positives);
            if (hag_curr->clause != NULL) {
                free(hag_curr->clause);
                hag_curr->clause = NULL;
            }
            HASH_DELETE(hag, hml_curr->ag, hag_curr);
            free(hag_curr);
        }
        HASH_CLEAR(hag, hml_curr->ag);
        HASH_DELETE(hml, *hash, hml_curr);
        free(hml_curr);
    }

    HASH_CLEAR(hml, *hash);
    free(*hash);
    *hash = NULL;
}

void free_learnt_clause_hash() {
    hml_pclauses *aux = NULL;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;  

    aux = learnt_clauses;
    while (aux != NULL) {
        hmc = aux->max;
        while (hmc != NULL) {
            hsc = hmc->size;
            while (hsc != NULL) {
                hsc->list = delete_clauses_list(hsc->list);
                if (hsc->list == NULL) {
                    HASH_DELETE(hsize,hmc->size,hsc);
                    free(hsc);
                    hsc = NULL;
                }
                else hsc=hsc->hsize.next;
            }
            if (hmc->size == NULL) {
                HASH_DELETE(hmax,aux->max,hmc);
                free(hmc);
                hmc = NULL;
            }
            else hmc=hmc->hmax.next;
        }
        if (aux->max == NULL) {
            HASH_DELETE(hml,learnt_clauses,aux);
            delete_clauses_list(aux->list_selection);
            free(aux);
            aux = NULL;
        }
        else aux=aux->hml.next;
    }
    HASH_CLEAR(hml,learnt_clauses);
    free(learnt_clauses);
    learnt_clauses=NULL;
}

void free_cnf_clauses() {
    if (modalliterals!=NULL) free_cnf_modal_hash(&modalliterals);
    /*if (learnt_clauses!=NULL) free_learnt_clause_hash();*/
}

// Check if lit appears on literals return 1
// Check if -lit appears on literals return -1
// Check if CFALSE appears on literals return -2
int isThisClauseSuitableForSat(int lit, literalslist* literals) {
    if(literals == NULL) return 0;
    literalslist *laux = literals;
    int find = 0;
    // Check if neg_int is CTRUE 
    if(lit == MAX_INT) find = 1;
    while(laux != NULL) {
        if(lit == laux->literal) find = 1;
        if(lit == -laux->literal) return -1;
        if(laux->literal == -MAX_INT) return -2;
        laux = laux->next;
    }

    return find;
}

clauseslist* insert_learnt_clause_list(tclause *clause, clauseslist *l) {
    clauseslist *new = malloc(sizeof(clauseslist));
    new->clause_number = clause->number;
    new->clause_pointer = clause;
    new->next = l;
    l = new;
    return l;
}

int clause_in_learnt_list(tclause *c, clauseslist *l) {
    while (l != NULL) {
        tclause *clause = l->clause_pointer;
        if (same(c,clause))
            return l->clause_number;
        else l = l->next;
    }
    return 0;
}

int check_repeated_learnt (tclause *clause, hml_pclauses *set) {
    int repeated = 0;
    hml_pclauses *hml = NULL;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;

    HASH_FIND(hml,set,&(clause->modal_level),sizeof(int),hml);
    if (hml != NULL) {
        HASH_FIND(hmax,hml->max,&(clause->max_literal),sizeof(int),hmc);
        if (hmc != NULL) {
            HASH_FIND(hsize,hmc->size,&(clause->size),sizeof(int),hsc);
            if (hsc != NULL) {
                repeated = clause_in_learnt_list(clause,hsc->list);
            }
        }
    }
    return repeated;
}

void insert_learnt_clause(tclause **clause) {
    tclause *aux = *clause;

    hml_pclauses *hml = NULL;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;

    HASH_FIND(hml,learnt_clauses,&(aux->modal_level),sizeof(int),hml);
    if (hml == NULL) {
        hml_pclauses *new1 = malloc(sizeof(*new1));
        new1->ml = aux->modal_level;
        new1->max = NULL;
        new1->list_selection = NULL;
        new1->last_selection = NULL;
        HASH_ADD(hml,learnt_clauses,ml,sizeof(int),new1);
        hml = new1;
    }

    HASH_FIND(hmax,hml->max,&(aux->max_literal),sizeof(int),hmc);
    if (hmc == NULL) {
        hmaxclauses *new2 = malloc(sizeof(*new2));
        new2->max = aux->max_literal;
        new2->size = NULL;
        HASH_ADD(hmax,hml->max,max,sizeof(int),new2);
        hmc = new2;
    }
    HASH_FIND(hsize,hmc->size,&(aux->size),sizeof(int),hsc);
    if (hsc == NULL) {
        hsizeclauses *new3 = malloc(sizeof(*new3));
        new3->size = aux->size;
        new3->list = NULL;
        HASH_ADD(hsize,hmc->size,size,sizeof(int),new3);
        hsc = new3;
    }
    hsc->list = insert_learnt_clause_list(aux, hsc->list);
}

void learn_clause_from_conflict(char* conflict_clause, int ml, int dia){
    char *strliteral;
    strliteral = strtok(conflict_clause, " ");
    int size = 1, max = 0, min = MAX_INT, lit;

    // Build new clause from conflict
    literalslist *iterator = NULL, *new_clause_literals = NULL;
    new_clause_literals = malloc(sizeof(literalslist));
    lit = atoi(strliteral);
    lit = (MAX_INT - abs(lit) + 1) * sign(lit);
    if(abs(lit) > max) max = lit;
    if(abs(lit) < min) min = lit;
    new_clause_literals->literal = lit;
    new_clause_literals->next = NULL;
    iterator = new_clause_literals;
    strliteral = strtok(NULL, " ");

    while(strliteral != NULL) {
        lit = atoi(strliteral);
        if(lit != 0) {
            lit = (MAX_INT - abs(lit) + 1) * sign(lit);
            if(abs(lit) > max) max = lit;
            if(abs(lit) < min) min = lit;
            iterator->next = malloc(sizeof(literalslist));
            iterator = iterator->next;
            iterator->literal = lit;
            iterator->next = NULL;
            size++;
        }
        strliteral = strtok(NULL, " ");
    }

    literalslist *with_dia = malloc(sizeof(literalslist));
    with_dia->literal = (MAX_INT - abs(dia) + 1) * sign(dia) * -1;
    with_dia->next = new_clause_literals;
    new_clause_literals = with_dia;

    tclause *newclause = malloc(sizeof(tclause));
    newclause->number = numclause++;
    newclause->type = LITERAL;
    newclause->size = size;
    newclause->modal_level = ml;
    newclause->max_literal = max;
    newclause->min_literal = min;
    newclause->left = CTRUE;
    newclause->agent = 0;
    newclause->propagated = 0;

    newclause->subtype = BOX; /* Change this at somepoint */
    newclause->numdiamond = BOX; /* Change this at somepoint */

    newclause->right = simplify_clause(new_clause_literals);
    newclause->size = size_clause(newclause->right);
    newclause->max_literal = max_lit_clause(newclause->right);
    newclause->min_literal = newclause->right->literal;
    newclause->class = classify_clause(newclause->right);

    /*newclause->right = negated_literals;*/
    newclause->just = malloc(sizeof(justification));
    newclause->just->rule = SAT_SOLVER;
    newclause->just->literals = NULL;
    newclause->just->parents = NULL;
    newclause->deleted = NULL;

    int repeated = 0;
    if(learnt_clauses != NULL) {
        repeated = check_repeated_learnt(newclause, learnt_clauses);
    }
    if(!repeated) {
        insert_learnt_clause(&newclause);
        insert_clause_prop(&newclause, &l_sos);
    }
    else {
        /*free_disjunction(new_clause_literals);*/
        free(newclause->just);
        free(newclause);
        newclause = NULL;
    }
}

void learn_clause_with_modal_literals(literalslist *pos_lits, int neg_literal, int ml) {
    literalslist *pos_lits_iterator = pos_lits;
    literalslist *negated_literals = NULL, *aux = NULL;
    int lit;

    negated_literals = malloc(sizeof(literalslist));
    lit = -neg_literal;
    negated_literals->literal = lit;
    negated_literals->next = NULL;

    while (pos_lits_iterator != NULL) {
        aux = malloc(sizeof(literalslist));
        lit = pos_lits_iterator->literal;
        lit = (MAX_INT - abs(lit) + 1)*sign(lit);
        aux->literal = -lit;
        aux->next = negated_literals;
        negated_literals = aux;
        pos_lits_iterator = pos_lits_iterator->next;
    }

    tclause *newclause = malloc(sizeof(tclause));
    newclause->number = numclause++;
    newclause->type = LITERAL;
    newclause->subtype = BOX; 
    newclause->numdiamond = BOX;
    newclause->modal_level = ml;
    newclause->agent = 1;
    newclause->left = CTRUE;
    newclause->propagated = 0;
    newclause->right = simplify_clause(negated_literals);
    newclause->size = size_clause(newclause->right);
    newclause->max_literal = max_lit_clause(newclause->right);
    newclause->min_literal = newclause->right->literal;
    newclause->class = classify_clause(newclause->right);

    newclause->just = malloc(sizeof(justification));
    newclause->just->rule = SAT_SOLVER;
    newclause->just->literals = NULL;
    newclause->just->parents = NULL;
    newclause->deleted = NULL;

    int repeated = 0;
    if(learnt_clauses != NULL) {
        repeated = check_repeated_learnt(newclause, learnt_clauses);
    }
    if(!repeated) {
        insert_learnt_clause(&newclause);
        insert_clause_prop(&newclause, &l_sos);
    }
    else {
        free_disjunction(negated_literals);
        free(newclause->just);
        free(newclause);
        newclause = NULL;
    }
}

void learn_clauses_every_modal_level() {
    hml_propclauses *hml_literals = NULL;
    if(propliterals != NULL) {
        for(hml_literals = propliterals; hml_literals != NULL; hml_literals = hml_literals->hml.next) {
            learn_clauses_at_modal_level_with_clauses(hml_literals->clauses, hml_literals->ml);
        }
    }
}

void learn_clauses_at_modal_level (int ml){
    hml_propclauses *prop_clauses_hash_at_ml;
    HASH_FIND(hml, propliterals, &ml, sizeof(int), prop_clauses_hash_at_ml);
    if(prop_clauses_hash_at_ml == NULL) return;
    char* propclauses = prop_clauses_hash_at_ml->clauses;

    learn_clauses_at_modal_level_with_clauses(propclauses, ml);
}

void learn_clauses_at_modal_level_with_clauses(char* propclauses, int ml) {
    hml_cnfliterals *hml = NULL;
    hag_cnfliterals *hag = NULL;
    literalslist *cnf_neg_lits = NULL;
    literalslist *cnf_pos_lits = NULL;
    char *final_modal_clause = NULL, *neg = NULL, *posclausestr = NULL;
    int neg_int, shouldAddClause = 0;

    int modal_level_of_modal_literals = ml - 1;

    HASH_FIND(hml, modalliterals, &modal_level_of_modal_literals, sizeof(int), hml);

    if (hml == NULL) return;
    for (hag = hml->ag; hag != NULL; hag = hag->hag.next) {
        cnf_neg_lits = hag->negatives;
        cnf_pos_lits = hag->positives;
        posclausestr = hag->clause;
        while (cnf_neg_lits != NULL) {
            neg_int = cnf_neg_lits->literal;
            shouldAddClause = isThisClauseSuitableForSat(neg_int, cnf_pos_lits);
            if(shouldAddClause < 0) continue;
            if(!shouldAddClause) {
                neg = itoa(neg_int);
                neg = realloc(neg, strlen(neg) + SPACE);
                final_modal_clause = malloc(strlen(posclausestr) + strlen(neg) + SPACE + CNF_EOL);
                sprintf(final_modal_clause, "%s%s 0\n", posclausestr, neg);
            }
            else {
                final_modal_clause = malloc(strlen(posclausestr) + SPACE + CNF_EOL);
                sprintf(final_modal_clause, "%s", posclausestr);
            }
            build_command(propclauses, final_modal_clause);
            free(final_modal_clause);
            final_modal_clause = NULL;

            //Call MiniSat
            FILE *fp;
            char path[1035];
            fp = popen(cmd, "r");
            if (fp == NULL) {
                printf("Failed to run command\n");
                exit(1);
            }

            while (fgets(path, sizeof(path) - 1, fp) != NULL) {
                if(path[0] == 'S') break;
                if(path[0] == 'U') {
                    neg_int = (MAX_INT - abs(neg_int) + 1)*sign(neg_int);
                    learn_clause_with_modal_literals(cnf_pos_lits, neg_int, ml);
                }
                else {
                    learn_clause_from_conflict(path, ml, neg_int);
                }
            }
            free(neg);
            neg = NULL;
            pclose(fp);
            free(cmd);
            cmd = NULL;
            cnf_neg_lits = cnf_neg_lits->next;
        }
    }
}
